/*
 *  TOPPERS/SSP Kernel
 *      Smallest Set Profile Kernel
 *
 *  Copyright (C) 2008 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2010 by Meika Sugimoto
 *  Copyright (C) 2018, 2019 by Naoki Saito
 *             Nagoya Municipal Industrial Research Institute, JAPAN
 *
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: core_config.c 384 2019-04-16 11:01:09Z nmir-saito $
 */

/*
 *  コア依存モジュール（ARMv8-A用）
 */

#include "kernel_impl.h"
#include <t_syslog.h>

/* 割込みネスト数の管理 */
uint8_t intnest;

/*
 *  コア依存の初期化
 */
void
core_initialize(void)
{
    /* 割込みネスト数の初期化 */
    intnest = 0u;
}

/*
 *  コア依存の終了処理
 */
void
core_terminate(void)
{
    while(1)
    ;
}

/*
 *  割込みハンドラの呼び出し
 */
void
call_int_handler(uint32_t intno)
{
    intnest++;

    i_unlock_cpu();

    // ハンドラの呼び出し
    inh_table[intno]();

    i_lock_cpu();

    intnest--;
}

/*
 *  カーネル管理のCPU例外ハンドラの呼び出し
 */
void
call_exc_handler(void *p_excinf, uint32_t excno)
{
    uint64_t spsr = *(((uint64_t*)p_excinf) + 4);

    intnest++;

    /*
     *  プロセッサの状態をCPU例外発生前の状態にする
     */
    if((spsr & AARCH64_PSTATE_F) == 0) {
        aarch64_clear_daif_f();
    }

    if((spsr & AARCH64_PSTATE_I) == 0) {
        aarch64_clear_daif_i();
    }

    // ハンドラの呼び出し
    exc_table[excno](p_excinf);

    /*
     *  プロセッサの状態を戻す
     */
    if((spsr & AARCH64_PSTATE_F) == 0) {
        aarch64_set_daif_f();
    }

    if((spsr & AARCH64_PSTATE_I) == 0) {
        aarch64_set_daif_i();
    }

    intnest--;
}

/*
 *  FIQハンドラの呼び出し
 *    全ての例外がマスクされた状態で呼び出される
 *    例外要因の判別も含めここ(あるいはここから呼び出されるルーチン)で行う
 */
void
call_fiq_handler(void)
{
}


#ifndef OMIT_DEFAULT_EXC_HANDLER
/*
 *  登録されていない例外が発生すると呼び出される
 */
void
default_exc_handler(void *p_excinf)
{
    uint64_t ipm = *(((uint64_t*)p_excinf) + 0);
    uint64_t excno    = *(((uint64_t*)p_excinf) + 1);
    uint64_t esr, pc, spsr, sp;

    esr      = *(((uint64_t*)p_excinf) + 2);
    pc       = *(((uint64_t*)p_excinf) + 3);
    spsr     = *(((uint64_t*)p_excinf) + 4);
    sp       = *(((uint64_t*)p_excinf) + 36);

    syslog(LOG_EMERG, "\nUnregistered CPU Exception occured.");
    syslog(LOG_EMERG, "EXCNO = %08x ESR = %08x iMASK = %08x ", excno, esr, ipm);
    syslog(LOG_EMERG, "PC=%08x%08x SP = %08x%08x SPSR = %08x",
        (uint32_t)((pc & UINT64_C(0xffffffff00000000))>>32), (uint32_t)(pc & UINT64_C(0xffffffff)),
        (uint32_t)((sp & UINT64_C(0xffffffff00000000))>>32), (uint32_t)(sp & UINT64_C(0xffffffff)),
        (uint32_t)spsr);

    target_exit();
}
#endif /* OMIT_DEFAULT_EXC_HANDLER */

#ifndef OMIT_DEFAULT_INT_HANDLER
/*
 *  未登録の割込みが発生した場合に呼び出される
 */
void
default_int_handler(void)
{
    syslog(LOG_EMERG, "\nUnregistered Interrupt occurs.");

    target_exit();
}
#endif /* OMIT_DEFAULT_INT_HANDLER */


